/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.scene.util.DrawSceneGraph;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
   Draw an interactive robot arm with shoulder, elbow, wrist, and finger joints.
   Make each part of the robot arm extendable.
<p>
   The tree for this scene is shown below.
<p>
   Remember that every position node in the tree contains a matrix,
   a model and a list of nested positions. The model may be empty,
   and the list of nested positions may also be empty, but the matrix
   cannot be "empty" (if you don't give it a value, then it is the
   identity matrix, I).
<p>
<pre>{@code
          Scene
         /     \
        /       \
  Camera   List<Position>
               |
               |
            Position
            /  |    \
           /   |     \
     Matrix    |     List<Position>
       RS      |           |
               |           |
              /        Position
             /         /   |   \
            /         /    |    \
           /     Matrix   /      List<Position>
          |       TRS    /             |
          |             /              |
          |            /            Position
          |           /             /  |   \
          |          /             /   |    \
          |         /        Matrix    |     List<Position>
          |        /          TRS      |           |
          |       /                   /            |
           \     /                   /          Position
            Model ------------------/           /  |   \
          armSegment                           /   |    \
                    \                    Matrix    |     List<Position>
                     \                    TRS     /             |
                      \                          /            empty
                       \------------------------/

</pre>
*/
public class RobotArm3 implements KeyListener, ComponentListener
{
   private double shoulderRotation = 0.0;
   private double    elbowRotation = 0.0;
   private double    wristRotation = 0.0;
   private double   fingerRotation = 0.0;

   private double shoulderLength = 0.4;
   private double    elbowLength = 0.3;
   private double    wristLength = 0.2;
   private double   fingerLength = 0.1;

   private final Scene scene;
   private Position shoulder_p;
   private Position elbow_p;
   private Position wrist_p;
   private Position finger_p;

   private boolean takeScreenshot = false;
   private int screenshotNumber = 0;

   private boolean useRenderer1 = true;

   private final JFrame jf;
   private final FrameBufferPanel fbp;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public RobotArm3()
   {
      // Create one Model that can be used
      // for each segment of the robot arm.
      final Vertex v0 = new Vertex(0, 0, 0);
      final Vertex v1 = new Vertex(1, 0, 0);
      final Model armSegment = new Model("arm_segment");
      armSegment.addVertex(v0, v1);
      armSegment.addPrimitive(new LineSegment(0, 1));
      ModelShading.setColor(armSegment, Color.blue);

      /*
         Create the scene graph.
      */
      scene = new Scene("Robot Arm 3",
                        Camera.projOrtho());

      shoulder_p = new Position(armSegment,
                                "shoulder",
                                Matrix.scale(shoulderLength,
                                             shoulderLength,
                                             1));
      scene.addPosition(shoulder_p);

      elbow_p = new Position(armSegment,
                             "elbow",
                             Matrix.translate(1, 0, 0)
                      .times(Matrix.scale(elbowLength/shoulderLength,
                                          elbowLength/shoulderLength,
                                          1)));
      shoulder_p.addNestedPosition(elbow_p);

      wrist_p = new Position(armSegment,
                             "wrist",
                             Matrix.translate(1, 0, 0)
                      .times(Matrix.scale(wristLength/elbowLength,
                                          wristLength/elbowLength,
                                          1)));
      elbow_p.addNestedPosition(wrist_p);

      finger_p = new Position(armSegment,
                              "finger",
                              Matrix.translate(1, 0, 0)
                       .times(Matrix.scale(fingerLength/wristLength,
                                           fingerLength/wristLength,
                                           1)));
      wrist_p.addNestedPosition(finger_p);

      DrawSceneGraph.draw(scene, "RobotArm_3");


      // Create a FrameBufferPanel that holds a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;
      fbp = new FrameBufferPanel(width, height, Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 10 - Robot Arm 3");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      print_help_message();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      if (e.isControlDown())
      {
         if (e.isShiftDown())
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^S (83)
            {
               shoulderLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^E (69)
            {
               elbowLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W) // ^W (87)
            {
               wristLength -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^F (70)
            {
               fingerLength -= 0.02;
            }
         }
         else // no shift key
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^s (93)
            {
               shoulderLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^e (69)
            {
               elbowLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W)  // ^w (87)
            {
               wristLength += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^f (70)
            {
               fingerLength += 0.02;
            }
         }
      }
   }

   @Override public void keyReleased(KeyEvent e){}

   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println( scene );
      }
      else if ('d' == c)
      {
         scene.debug = ! scene.debug;
         Clip.debug = scene.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('1' == c)
      {
         if (! useRenderer1)
         {
            useRenderer1 = true;
         }
         System.out.println("Using Pipeline 1.");
      }
      else if ('2' == c)
      {
         if (useRenderer1)
         {
            useRenderer1 = false;
         }
         System.out.println("Using Pipeline 2.");
      }
      else if ('c' == c)
      {
         // Change the solid random color of the robot arm.
         final Color color = ModelShading.randomColor();
         ModelShading.setColor(shoulder_p.model, color);
      }
      else if ('C' == c)
      {
         // Change the solid random color of each segment of the robot arm.
         ModelShading.setRandomColor(shoulder_p.model);
         ModelShading.setRandomColor(elbow_p.model);
         ModelShading.setRandomColor(wrist_p.model);
         ModelShading.setRandomColor(finger_p.model);
      }
      else if ('r' == c)
      {
         // Change the random color at each end of each segment of the robot arm.
         ModelShading.setRainbowPrimitiveColors(shoulder_p.model);
      }
      else if ('R' == c)
      {
         // Change the random color at each vertex of the robot arm.
         final Color c1 = ModelShading.randomColor();
         final Color c2 = ModelShading.randomColor();
         shoulder_p.model.colorList.clear();
         shoulder_p.model.addColor(c1, c2);
         shoulder_p.model.getPrimitive(0).setColorIndices(0, 1);
      }
      else if ('=' == c)
      {
         shoulderRotation = 0.0;
            elbowRotation = 0.0;
            wristRotation = 0.0;
           fingerRotation = 0.0;

           shoulderLength = 0.4;
              elbowLength = 0.3;
              wristLength = 0.2;
             fingerLength = 0.1;
      }
      else if ('s' == c)
      {
         shoulderRotation += 2.0;
      }
      else if ('S' == c)
      {
         shoulderRotation -= 2.0;
      }
      else if ('e' == c)
      {
         elbowRotation += 2.0;
      }
      else if ('E' == c)
      {
         elbowRotation -= 2.0;
      }
      else if ('w' == c)
      {
         wristRotation += 2.0;
      }
      else if ('W' == c)
      {
         wristRotation -= 2.0;
      }
      else if ('f' == c)
      {
         fingerRotation += 2.0;
      }
      else if ('F' == c)
      {
         fingerRotation -= 2.0;
      }
      else if ('+' == c)
      {
         takeScreenshot = true;
      }

      // Update the nested matrices for the sub models.
      shoulder_p = shoulder_p.transform(
                     Matrix.rotateZ(shoulderRotation)
              .times(Matrix.scale(shoulderLength,
                                  shoulderLength,
                                  1)));
      scene.setPosition(0, shoulder_p);

      elbow_p = elbow_p.transform(
                     Matrix.translate(1, 0, 0)
              .times(Matrix.rotateZ(elbowRotation))
              .times(Matrix.scale(elbowLength/shoulderLength,
                                  elbowLength/shoulderLength,
                                  1)));
      shoulder_p.setNestedPosition(0, elbow_p);

      wrist_p = wrist_p.transform(
                     Matrix.translate(1, 0, 0)
             .times(Matrix.rotateZ(wristRotation))
             .times(Matrix.scale(wristLength/elbowLength,
                                 wristLength/elbowLength,
                                 1)));
      elbow_p.setNestedPosition(0, wrist_p);

      finger_p = finger_p.transform(
                     Matrix.translate(1, 0, 0)
              .times(Matrix.rotateZ(fingerRotation))
              .times(Matrix.scale(fingerLength/wristLength,
                                  fingerLength/wristLength,
                                  1)));
      wrist_p.setNestedPosition(0, finger_p);

      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      if (useRenderer1)
      {
         Pipeline.render(scene, fb);
      }
      else
      {
         Pipeline2.render(scene, fb);
      }
      if (takeScreenshot)
      {
         fb.dumpFB2File(String.format("Screenshot%03d.png", screenshotNumber), "png");
         ++screenshotNumber;
         takeScreenshot = false;
      }
      fbp.repaint();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg = fbp.getFrameBuffer().getBackgroundColorFB();
      final FrameBuffer fb = new FrameBuffer(w, h, bg);
      fbp.setFrameBuffer(fb);
      if (useRenderer1)
      {
         Pipeline.render(scene, fb);
      }
      else
      {
         Pipeline2.render(scene, fb);
      }
      fbp.repaint();
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'Alt-d' key combination to print the Scene data structure.");
      System.out.println("Use the '1' and '2' keys to switch between the two renderers.");
      System.out.println("Use the 'c' key to change the random solid arm color.");
      System.out.println("Use the 'C' key to randomly change arm segment colors.");
      System.out.println("Use the 'r' key to randomly change arm segment end colors.");
      System.out.println("Use the 'R' key to randomly change arm hinge colors.");
      System.out.println("Use the s/S keys to rotate the arm at the shoulder.");
      System.out.println("Use the e/E keys to rotate the arm at the elbow.");
      System.out.println("Use the w/W keys to rotate the arm at the wrist.");
      System.out.println("Use the f/F keys to rotate the arm at the finger.");
      System.out.println("Use the ^s/^S keys to extend the length of the arm at the shoulder.");
      System.out.println("Use the ^e/^E keys to extend the length of the arm at the elbow.");
      System.out.println("Use the ^w/^W keys to extend the length of the arm at the wrist.");
      System.out.println("Use the ^f/^F keys to extend the length of the arm at the finger.");
      System.out.println("Use the '=' key to reset the robot arm.");
      System.out.println("Use the '+' key to save a \"screenshot\" of the framebuffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new RobotArm3()
      );
   }
}
