/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.ModelShading;
import renderer.scene.util.DrawSceneGraph;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;
import javax.swing.JFrame;
import java.awt.BorderLayout;
import java.awt.event.KeyListener;
import java.awt.event.KeyEvent;
import java.awt.event.ComponentListener;
import java.awt.event.ComponentEvent;

/**
   Draw two interactive robot arms with
   shoulder, elbow, wrist, and finger joints.
<p>
   Here is a simplified version of this program's scene graph.
<p>
<pre>{@code
                     Scene
                    /     \
                   /       \
             Camera   List<Position>
                           /   \
                   /------/     \------\
           Position                     Position
           / |   \                       / |   \
          /  |    \                     /  |    \
    Matrix   |  List<Position>    Matrix   |   List<Position>
     TRS     |     |               TRS     |      |
             |     |                       |      |
            /  Position                   /   Position
           /     / |  \                  /     /  |  \
          /     /  |   \                /     /   |   \
         / Matrix  |    \              / Matrix   |    \
        /   TRS   /    Position       |   TRS    /   Position
        \        /      / |  \        |         /    /  |  \
         \      /      /  |   \       |        /    /   |   \
          \    /  Matrix  |    \      |       / Matrix  |    \
           \   \   TRS   /   Position |      |   TRS   /   Position
            \   \       /      /  |   |      |        /     /    /
             \   \     /      /   |   |      |       /     /    /
              \   \   /  Matrix   |   |     /       /  Matrix  /
               \   \  \   TRS     |   |    /       /    TRS   /
                \   \  \          |   |   /       /          /
                 \   \  \------\  |   |  /       /          /
                  \   \---------\ |   | /-------/          /
                   \-------------\|   |/------------------/
                                  Model
                                armSegment
</pre>
*/
public class RobotArms2 implements KeyListener, ComponentListener
{
   private double xTranslation1 = 0.0;
   private double yTranslation1 = 0.5;
   private double xTranslation2 =  0.0;
   private double yTranslation2 = -0.5;

   private double shoulderRotation1 = 0.0;
   private double shoulderRotation2 = 0.0;
   private double    elbowRotation1 = 0.0;
   private double    elbowRotation2 = 0.0;
   private double    wristRotation1 = 0.0;
   private double    wristRotation2 = 0.0;
   private double   fingerRotation1 = 0.0;
   private double   fingerRotation2 = 0.0;

   private double shoulderLength1 = 0.4;
   private double shoulderLength2 = 0.4;
   private double    elbowLength1 = 0.3;
   private double    elbowLength2 = 0.3;
   private double    wristLength1 = 0.2;
   private double    wristLength2 = 0.2;
   private double   fingerLength1 = 0.1;
   private double   fingerLength2 = 0.1;

   private final Scene scene;
   private Position shoulder1_p;
   private Position shoulder2_p;
   private Position elbow1_p;
   private Position elbow2_p;
   private Position wrist1_p;
   private Position wrist2_p;
   private Position finger1_p;
   private Position finger2_p;

   private boolean takeScreenshot = false;
   private int screenshotNumber = 0;

   private boolean useRenderer1 = true;

   private final JFrame jf;
   private final FrameBufferPanel fbp;

   /**
      This constructor instantiates the Scene object
      and initializes it with appropriate geometry.
      Then this constructor instantiates the GUI.
   */
   public RobotArms2()
   {
      // Create one Model that can be used
      // for each segment of the robot arms.
      final Vertex v0 = new Vertex(0, 0, 0);
      final Vertex v1 = new Vertex(1, 0, 0);
      final Model armSegment = new Model("arm_segment");
      armSegment.addVertex(v0, v1);
      armSegment.addPrimitive(new LineSegment(0, 1));
      ModelShading.setColor(armSegment, Color.blue);

      /*
         Create the scene graph.
      */
      scene = new Scene("Robot Arms 2",
                        Camera.projOrtho());

      /*
         Create two robot arms.
      */
      // First arm.
      shoulder1_p = new Position(armSegment,
                                 "arm_1",
                                 Matrix.translate(xTranslation1,
                                                  yTranslation1,
                                                  0)
                         .times(Matrix.rotateZ(shoulderRotation1))
                         .times(Matrix.scale(shoulderLength1,
                                             shoulderLength1,
                                             1)));
      scene.addPosition(shoulder1_p);

      elbow1_p = new Position(armSegment,
                              "elbow_1",
                              Matrix.translate(1, 0, 0)
                       .times(Matrix.rotateZ(elbowRotation1))
                       .times(Matrix.scale(elbowLength1/shoulderLength1,
                                           elbowLength1/shoulderLength1,
                                           1)));
      shoulder1_p.addNestedPosition(elbow1_p);

      wrist1_p = new Position(armSegment,
                              "wrist_1",
                              Matrix.translate(1, 0, 0)
                       .times(Matrix.rotateZ(wristRotation1))
                       .times(Matrix.scale(wristLength1/elbowLength1,
                                           wristLength1/elbowLength1,
                                           1)));
      elbow1_p.addNestedPosition(wrist1_p);

      finger1_p = new Position(armSegment,
                               "finger_1",
                               Matrix.translate(1, 0, 0)
                        .times(Matrix.rotateZ(fingerRotation1))
                        .times(Matrix.scale(fingerLength1/wristLength1,
                                            fingerLength1/wristLength1,
                                            1)));
      wrist1_p.addNestedPosition(finger1_p);

      // Second arm.
      shoulder2_p = new Position(armSegment,
                                 "arm_2",
                                 Matrix.translate(xTranslation2,
                                                  yTranslation2,
                                                  0)
                         .times(Matrix.rotateZ(shoulderRotation2))
                         .times(Matrix.scale(shoulderLength2,
                                             shoulderLength2,
                                             1)));
      scene.addPosition(shoulder2_p);

      elbow2_p = new Position(armSegment,
                              "elbow_2",
                              Matrix.translate(1, 0, 0)
                       .times(Matrix.rotateZ(elbowRotation2))
                       .times(Matrix.scale(elbowLength2/shoulderLength2,
                                           elbowLength2/shoulderLength2,
                                           1)));
      shoulder2_p.addNestedPosition(elbow2_p);

      wrist2_p = new Position(armSegment,
                              "wrist_2",
                              Matrix.translate(1, 0, 0)
                       .times(Matrix.rotateZ(wristRotation2))
                       .times(Matrix.scale(wristLength2/elbowLength2,
                                           wristLength2/elbowLength2,
                                           1)));
      elbow2_p.addNestedPosition(wrist2_p);

      finger2_p = new Position(armSegment,
                               "finger_2",
                               Matrix.translate(1, 0, 0)
                        .times(Matrix.rotateZ(fingerRotation2))
                        .times(Matrix.scale(fingerLength2/wristLength2,
                                            fingerLength2/wristLength2,
                                            1)));
      wrist2_p.addNestedPosition(finger2_p);

      DrawSceneGraph.draw(scene, "RobotArms_2");


      // Create a FrameBufferPanel that holds a FrameBuffer.
      final int width  = 1024;
      final int height = 1024;
      fbp = new FrameBufferPanel(width, height, Color.black);

      // Create a JFrame that will hold the FrameBufferPanel.
      jf = new JFrame("Renderer 10 - Robot Arms 2");
      jf.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
      jf.getContentPane().add(fbp, BorderLayout.CENTER);
      jf.pack();
      jf.setLocationRelativeTo(null);
      jf.setVisible(true);

      // Create event handler objects for events from the JFrame.
      jf.addKeyListener(this);
      jf.addComponentListener(this);

      print_help_message();
   }


   // Implement the KeyListener interface.
   @Override public void keyPressed(KeyEvent e)
   {
      //System.out.println( e );

      if (e.isControlDown())
      {
         if (e.isShiftDown())
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^S (83)
            {
               shoulderLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^E (69)
            {
               elbowLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W) // ^W (87)
            {
               wristLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^F (70)
            {
               fingerLength1 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_B) // ^B
            {
               shoulderLength2 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Q) // ^Q
            {
               elbowLength2 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Z) // ^Z
            {
               wristLength2 -= 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_A) // ^A
            {
               fingerLength2 -= 0.02;
            }
         }
         else // no shift key
         {
            if (e.getKeyCode() == KeyEvent.VK_S) // ^s (93)
            {
               shoulderLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_E) // ^e (69)
            {
               elbowLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_W)  // ^w (87)
            {
               wristLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_F) // ^f (70)
            {
               fingerLength1 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_B) // ^b
            {
               shoulderLength2 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Q) // ^q
            {
               elbowLength2 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_Z)  // ^z
            {
               wristLength2 += 0.02;
            }
            else if (e.getKeyCode() == KeyEvent.VK_A) // ^a
            {
               fingerLength2 += 0.02;
            }
         }
      }
   }

   @Override public void keyReleased(KeyEvent e){}

   @Override public void keyTyped(KeyEvent e)
   {
      //System.out.println( e );

      final char c = e.getKeyChar();
      if ('h' == c)
      {
         print_help_message();
         return;
      }
      else if ('d' == c && e.isAltDown())
      {
         System.out.println();
         System.out.println( scene );
      }
      else if ('d' == c)
      {
         scene.debug = ! scene.debug;
         Clip.debug = scene.debug;
      }
      else if ('D' == c)
      {
         Rasterize.debug = ! Rasterize.debug;
      }
      else if ('1' == c)
      {
         if (! useRenderer1)
         {
            useRenderer1 = true;
         }
         System.out.println("Using Pipeline 1.");
      }
      else if ('2' == c)
      {
         if (useRenderer1)
         {
            useRenderer1 = false;
         }
         System.out.println("Using Pipeline 2.");
      }
      else if ('c' == c)
      {
         // Change the solid random color of the robot arm.
         final Color color = ModelShading.randomColor();
         ModelShading.setColor(shoulder1_p.model, color);
         ModelShading.setColor(elbow1_p.model, color);
         ModelShading.setColor(wrist1_p.model, color);
         ModelShading.setColor(finger1_p.model, color);
      }
      else if ('C' == c)
      {
         // Change the solid random color of each segment of the robot arm.
         ModelShading.setRandomColor(shoulder1_p.model);
         ModelShading.setRandomColor(elbow1_p.model);
         ModelShading.setRandomColor(wrist1_p.model);
         ModelShading.setRandomColor(finger1_p.model);
      }
      else if ('r' == c)
      {
         // Change the random color at each end of each segment of the robot arm.
         ModelShading.setRainbowPrimitiveColors(shoulder1_p.model);
         ModelShading.setRainbowPrimitiveColors(elbow1_p.model);
         ModelShading.setRainbowPrimitiveColors(wrist1_p.model);
         ModelShading.setRainbowPrimitiveColors(finger1_p.model);
      }
      else if ('R' == c)
      {
         // Change the random color at each vertex of the robot arm.
         final Color c1 = ModelShading.randomColor();
         final Color c2 = ModelShading.randomColor();
         final Color c3 = ModelShading.randomColor();
         final Color c4 = ModelShading.randomColor();
         final Color c5 = ModelShading.randomColor();
         shoulder1_p.model.colorList.clear();
            elbow1_p.model.colorList.clear();
            wrist1_p.model.colorList.clear();
           finger1_p.model.colorList.clear();
         shoulder1_p.model.addColor(c1, c2);
            elbow1_p.model.addColor(c2, c3);
            wrist1_p.model.addColor(c3, c4);
           finger1_p.model.addColor(c4, c5);
         shoulder1_p.model.getPrimitive(0).setColorIndices(0, 1);
            elbow1_p.model.getPrimitive(0).setColorIndices(0, 1);
            wrist1_p.model.getPrimitive(0).setColorIndices(0, 1);
           finger1_p.model.getPrimitive(0).setColorIndices(0, 1);
      }
      else if ('=' == c)
      {
         xTranslation1 =  0.0;
         yTranslation1 =  0.5;
         xTranslation2 =  0.0;
         yTranslation2 = -0.5;

        shoulderRotation1 = 0.0;
        shoulderRotation2 = 0.0;
           elbowRotation1 = 0.0;
           elbowRotation2 = 0.0;
           wristRotation1 = 0.0;
           wristRotation2 = 0.0;
          fingerRotation1 = 0.0;
          fingerRotation2 = 0.0;

          shoulderLength1 = 0.4;
          shoulderLength2 = 0.4;
             elbowLength1 = 0.3;
             elbowLength2 = 0.3;
             wristLength1 = 0.2;
             wristLength2 = 0.2;
            fingerLength1 = 0.1;
            fingerLength2 = 0.1;
      }
      else if ('x' == c)
      {
         xTranslation1 += 0.02;
      }
      else if ('X' == c)
      {
         xTranslation1 -= 0.02;
      }
      else if ('y' == c)
      {
         yTranslation1 += 0.02;
      }
      else if ('Y' == c)
      {
         yTranslation1 -= 0.02;
      }
      else if ('u' == c)
      {
         xTranslation2 += 0.02;
      }
      else if ('U' == c)
      {
         xTranslation2 -= 0.02;
      }
      else if ('v' == c)
      {
         yTranslation2 += 0.02;
      }
      else if ('V' == c)
      {
         yTranslation2 -= 0.02;
      }
      else if ('s' == c)
      {
         shoulderRotation1 += 2.0;
      }
      else if ('S' == c)
      {
         shoulderRotation1 -= 2.0;
      }
      else if ('e' == c)
      {
         elbowRotation1 += 2.0;
      }
      else if ('E' == c)
      {
         elbowRotation1 -= 2.0;
      }
      else if ('w' == c)
      {
         wristRotation1 += 2.0;
      }
      else if ('W' == c)
      {
         wristRotation1 -= 2.0;
      }
      else if ('f' == c)
      {
         fingerRotation1 += 2.0;
      }
      else if ('F' == c)
      {
         fingerRotation1 -= 2.0;
      }
      else if ('b' == c)
      {
         shoulderRotation2 += 2.0;
      }
      else if ('B' == c)
      {
         shoulderRotation2 -= 2.0;
      }
      else if ('q' == c)
      {
         elbowRotation2 += 2.0;
      }
      else if ('Q' == c)
      {
         elbowRotation2 -= 2.0;
      }
      else if ('z' == c)
      {
         wristRotation2 += 2.0;
      }
      else if ('Z' == c)
      {
         wristRotation2 -= 2.0;
      }
      else if ('a' == c)
      {
         fingerRotation2 += 2.0;
      }
      else if ('A' == c)
      {
         fingerRotation2 -= 2.0;
      }
      else if ('+' == c)
      {
         takeScreenshot = true;
      }

      // Update the nested matrices for the sub models.
      // First arm.
      shoulder1_p = shoulder1_p.transform(
                        Matrix.translate(xTranslation1,
                                         yTranslation1,
                                         0)
                 .times(Matrix.rotateZ(shoulderRotation1))
                 .times(Matrix.scale(shoulderLength1,
                                    shoulderLength1,
                                    1)));
      scene.setPosition(0, shoulder1_p);

      elbow1_p = elbow1_p.transform(
                     Matrix.translate(1, 0, 0)
              .times(Matrix.rotateZ(elbowRotation1))
              .times(Matrix.scale(elbowLength1/shoulderLength1,
                                 elbowLength1/shoulderLength1,
                                 1)));
      shoulder1_p.setNestedPosition(0, elbow1_p);

      wrist1_p = wrist1_p.transform(
                     Matrix.translate(1, 0, 0)
              .times(Matrix.rotateZ(wristRotation1))
              .times(Matrix.scale(wristLength1/elbowLength1,
                                 wristLength1/elbowLength1,
                                 1)));
      elbow1_p.setNestedPosition(0, wrist1_p);

      finger1_p = finger1_p.transform(
                      Matrix.translate(1, 0, 0)
               .times(Matrix.rotateZ(fingerRotation1))
               .times(Matrix.scale(fingerLength1/wristLength1,
                                  fingerLength1/wristLength1,
                                  1)));
      wrist1_p.setNestedPosition(0, finger1_p);

      // Second arm.
      shoulder2_p = shoulder2_p.transform(
                        Matrix.translate(xTranslation2,
                                         yTranslation2,
                                         -1)
                 .times(Matrix.rotateZ(shoulderRotation2))
                 .times(Matrix.scale(shoulderLength2,
                                    shoulderLength2,
                                     1)));
      scene.setPosition(1, shoulder2_p);

      elbow2_p = elbow2_p.transform(
                     Matrix.translate(1, 0, 0)
              .times(Matrix.rotateZ(elbowRotation2))
              .times(Matrix.scale(elbowLength2/shoulderLength2,
                                 elbowLength2/shoulderLength2,
                                 1)));
      shoulder2_p.setNestedPosition(0, elbow2_p);

      wrist2_p = wrist2_p.transform(
                     Matrix.translate(1, 0, 0)
              .times(Matrix.rotateZ(wristRotation2))
              .times(Matrix.scale(wristLength2/elbowLength2,
                                 wristLength2/elbowLength2,
                                 1)));
      elbow2_p.setNestedPosition(0, wrist2_p);

      finger2_p = finger2_p.transform(
                      Matrix.translate(1, 0, 0)
               .times(Matrix.rotateZ(fingerRotation2))
               .times(Matrix.scale(fingerLength2/wristLength2,
                                  fingerLength2/wristLength2,
                                  1)));
      wrist2_p.setNestedPosition(0, finger2_p);

      // Render again.
      final FrameBuffer fb = this.fbp.getFrameBuffer();
      fb.clearFB();
      if (useRenderer1)
      {
         Pipeline.render(scene, fb);
      }
      else
      {
         Pipeline2.render(scene, fb);
      }
      if (takeScreenshot)
      {
         fb.dumpFB2File(String.format("Screenshot%03d.png", screenshotNumber), "png");
         ++screenshotNumber;
         takeScreenshot = false;
      }
      fbp.repaint();
   }


   // Implement the ComponentListener interface.
   @Override public void componentMoved(ComponentEvent e){}
   @Override public void componentHidden(ComponentEvent e){}
   @Override public void componentShown(ComponentEvent e){}
   @Override public void componentResized(ComponentEvent e)
   {
      //System.out.println( e );
      /*
      System.out.printf("JFrame [w = %d, h = %d]: " +
                        "FrameBufferPanel [w = %d, h = %d].\n",
                        jf.getWidth(), jf.getHeight(),
                        fbp.getWidth(), fbp.getHeight());
      */
      // Get the new size of the FrameBufferPanel.
      final int w = fbp.getWidth();
      final int h = fbp.getHeight();

      // Create a new FrameBuffer that fits the FrameBufferPanel.
      final Color bg = fbp.getFrameBuffer().getBackgroundColorFB();
      final FrameBuffer fb = new FrameBuffer(w, h, bg);
      fbp.setFrameBuffer(fb);
      if (useRenderer1)
      {
         Pipeline.render(scene, fb);
      }
      else
      {
         Pipeline2.render(scene, fb);
      }
      fbp.repaint();
   }


   private void print_help_message()
   {
      System.out.println("Use the 'd' key to toggle debugging information on and off.");
      System.out.println("Use the 'Alt-d' key combination to print the Scene data structure.");
      System.out.println("Use the '1' and '2' keys to switch between the two renderers.");
      System.out.println("Use the 'c' key to change the random solid arm color.");
      System.out.println("Use the 'C' key to randomly change arm segment colors.");
      System.out.println("Use the 'r' key to randomly change arm segment end colors.");
      System.out.println("Use the 'R' key to randomly change arm hinge colors.");
      System.out.println();
      System.out.println("Use the s/S keys to rotate arm 1 at its shoulder.");
      System.out.println("Use the e/E keys to rotate arm 1 at its elbow.");
      System.out.println("Use the w/W keys to rotate arm 1 at its wrist.");
      System.out.println("Use the f/F keys to rotate arm 1 at its finger.");
      System.out.println();
      System.out.println("Use the b/B keys to rotate arm 2 at its shoulder.");
      System.out.println("Use the q/Q keys to rotate arm 2 at its elbow.");
      System.out.println("Use the z/Z keys to rotate arm 2 at its wrist.");
      System.out.println("Use the a/A keys to rotate arm 2 at its finger.");
      System.out.println();
      System.out.println("Use the ^s/^S keys to extend the length of arm 1 at its shoulder.");
      System.out.println("Use the ^e/^E keys to extend the length of arm 1 at its elbow.");
      System.out.println("Use the ^w/^W keys to extend the length of arm 1 at its wrist.");
      System.out.println("Use the ^f/^F keys to extend the length of arm 1 at its finger.");
      System.out.println();
      System.out.println("Use the ^b/^B keys to extend the length of arm 2 at its shoulder.");
      System.out.println("Use the ^q/^Q keys to extend the length of arm 2 at its elbow.");
      System.out.println("Use the ^z/^Z keys to extend the length of arm 2 at its wrist.");
      System.out.println("Use the ^a/^A keys to extend the length of arm 2 at its finger.");
      System.out.println();
      System.out.println("Use the x/X keys to translate arm 1 along the x-axis.");
      System.out.println("Use the y/Y keys to translate arm 1 along the y-axis.");
      System.out.println("Use the u/U keys to translate arm 2 along the x-axis.");
      System.out.println("Use the v/V keys to translate arm 2 along the y-axis.");
      System.out.println();
      System.out.println("Use the '=' key to reset the robot arms.");
      System.out.println("Use the '+' key to save a \"screenshot\" of the framebuffer.");
      System.out.println("Use the 'h' key to redisplay this help message.");
   }


   /**
      Create an instance of this class which has
      the affect of creating the GUI application.
   */
   public static void main(String[] args)
   {
      // We need to call the program's constructor in the
      // Java GUI Event Dispatch Thread, otherwise we get a
      // race condition between the constructor (running in
      // the main() thread) and the very first ComponentEvent
      // (running in the EDT).
      javax.swing.SwingUtilities.invokeLater(
         () -> new RobotArms2()
      );
   }
}
