/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This file defines a hierarchical scene out of a single instance of Model_1.
<p>
   The graph for this scene is shown below. Since two Position objects
   share a reference to the same Model_1 object, this graph is not a tree,
   it is a "directed acyclic graph" (DAG). In general, scene graphs can be
   DAGs since multiple Position objects can share a Model object, and
   Position objects can also share a (nested) Position object.
<p>
<pre>{@code
           Scene
          /     \
         /       \
    Camera     List<Position>
                     |
                     |
                  Position
                 /   |    \
                /    |     \
          Matrix   Model     List<Position>
            R     (empty)   /              \
                           /                \
                          /                  \
                   Position             Position
                    /      \            /      /
                   /        \          /      /
               Matrix        \      Matrix   /
                 I            \      TSR    /
                               \           /
                                \         /
                                  Model_1
}</pre>
*/
public class ShowModel_2a
{
   public static void main(String[] args)
   {
      // Create a single instance of Model_1.
      final Model m1 = new Model_1();
      ModelShading.setColor(m1, Color.red);

      // Create a matrix for model m1.
      final Matrix mat1 = Matrix.translate(1, -2-Math.sqrt(2), 0)
                  .times( Matrix.scale(0.5, 0.5, 1) )
                  .times( Matrix.rotateZ(-45) );

      final Scene scene = new Scene("ShowModel_2",
                                    Camera.projOrtho(-5, +5, -5, +5));

      // Create the top level Position.
      final Position p0 = new Position(new Model("empty"),
                                       "top");

      // Add the top level Position to the Scene.
      scene.addPosition(p0);

      // Add two nested Positions to the top level Position.
      final Position p1 = new Position(m1, "left", mat1);
      final Position p2 = new Position(m1, "right");
      p0.addNestedPosition(p1, p2);

      DrawSceneGraph.draw(scene, "SG_ShowModel_2a");
      DrawSceneGraphTopologicalSort.draw(scene, "SG_ShowModel_2a_TS");


      // Create a FrameBuffer to render our Scene into.
      final int width  = 1024;
      final int height = 1024;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.white);

      //PipelineLogger.debug = true;

      for (int i = 0; i <= 36; ++i)
      {
         scene.setPosition(0,
                  scene.getPosition(0).transform( Matrix.rotateZ(10*i) ));

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_ShowModel_2a_Frame%02d.ppm", i));
      }
   }
}
