/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   The DAG for this scene is shown below. There are two instances of
   Model_1 in the DAG and each instance has two paths in the DAG from
   the root Scene object to that instance. So Model_1 still ends up
   appearing in the scene in four places. But since there are two
   instances of Model_1, each instance can be given its own color.
   So we see two red and two blue instances of Model_1 in the scene.
<p>
<pre>{@code
           Scene
          /     \
         /       \
    Camera     List<Position>
                     |
                     |
                  Position
                 /   |    \
                /    |     \
          Matrix   Model     List<Position>
            R     (empty)   /              \
                           /                \
                          /                  \
                  Position                    Position
                 /   |    \                   /   |   \
                /    |     \                 /    |    \
          Matrix   Model List<Position>  Matrix Model   List<Position>
            T     (empty)          \      TR   (empty)  /
                                    \                  /
                                     \                /
                                      \              /
                                       \            /
                                        \          /
                                          Position
                                         /  |    \
                                        /   |     \
                                  Matrix  Model    List<Position>
                                    I    (empty)    /           \
                                                   /             \
                                                  /               \
                                           Position             Position
                                           /     |             /      |
                                          /      |            /       |
                                      Matrix   Model_1    Matrix    Model_1
                                        I      (red)       TSR      (blue)
}</pre>
*/
public class ShowModel_4
{
   public static void main(String[] args)
   {
      // Create two instances of Model_1.
      final Model m1 = new Model_1();
      final Model m2 = new Model_1();
      ModelShading.setColor(m1, Color.red);
      ModelShading.setColor(m2, Color.blue);

      // Create matricies for the nested Positions.
      final Matrix mat1 = Matrix.translate(-2, -2, 0);
      final Matrix mat2 = Matrix.translate(2, 2, 0)
                  .times( Matrix.rotateZ(180) );
      final Matrix mat5 = Matrix.translate(1, -2-Math.sqrt(2), 0)
                  .times( Matrix.scale(0.5, 0.5, 1) )
                  .times( Matrix.rotateZ(-45) );

      final Scene scene = new Scene("ShowModel_4",
                                    Camera.projOrtho(-8, +8, -8, +8));

      // Create the top level Position.
      final Position p0 = new Position(new Model("empty"), "top");

      // Add the top level Position to the Scene.
      scene.addPosition(p0);

      // Add two nested Positions to the top level Position.
      final Position p1 = new Position(new Model("empty"), "p1", mat1);
      final Position p2 = new Position(new Model("empty"), "p2", mat2);
      p0.addNestedPosition(p1, p2);

      // Add a reference to a Position p3 to each of Positions p1 and p2.
      final Position p3 = new Position(new Model("empty"), "p3");
      p1.addNestedPosition(p3);
      p2.addNestedPosition(p3);

      // Add two nested Positions to the Position p3.
      final Position p4 = new Position(m1, "p4");
      final Position p5 = new Position(m2, "p5", mat5);
      p3.addNestedPosition(p4, p5);

      DrawSceneGraph.draw(scene, "SG_ShowModel_4");
      DrawSceneGraphTopologicalSort.draw(scene, "SG_ShowModel_4_TS");


      // Create a FrameBuffer to render our Scene into.
      final int width  = 1024;
      final int height = 1024;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.white);

      //PipelineLogger.debug = true;

      for (int i = 0; i <= 36; ++i)
      {
         scene.setPosition(0,
                  scene.getPosition(0).transform( Matrix.rotateZ(10*i) ));

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_ShowModel_4_Frame%02d.ppm", i));
      }
   }
}
