/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This program creates a very complex hierarchical scene
   that draws 36 copies of Model_1.
<p>
   The DAG has a top level Position that holds nine nested
   positions. Each of the nine nested positions holds a
   reference to a sub-DAG that is copied from ShowModel_4.java.
   There are 36 paths in the DAG from the root Scene object to
   the single Model_1 object at the bottom of the DAG.
<p>
<pre>{@code
        Scene
        /   \
       /     \
  Camera   List<Position>
             |
             |
         Position
         /  |    \
        /   |     \
  Matrix  Model     List<Position>
    R    (empty)    /             \
                   /               \
                  /                 \
                 /                   \
                /                     \
               /                       \
         Pos Pos Pos Pos Pos Pos Pos Pos Pos
               \                      /
                \                    /
                 \                  /
                  \                /
                   \              /
                    \            /
                       Position
                       /  |    \
                      /   |     \
                Matrix  Model     List<Position>
                  I    (empty)  /                \
                               /                  \
                              /                    \
                       Position                    Position
                      /   |    \                   /   |   \
                     /    |     \                 /    |    \
               Matrix   Model  List<Position> Matrix Model   List<Position>
                 T     (empty)         \       TR   (empty)  /
                                        \                   /
                                         \                 /
                                          \               /
                                           \             /
                                            \           /
                                              Position
                                              /  |    \
                                             /   |     \
                                       Matrix  Model    List<Position>
                                         I    (empty)    /           \
                                                        /             \
                                                       /               \
                                                Position            Position
                                                /       \           /     /
                                               /         \         /     /
                                           Matrix         \    Matrix   /
                                             I             \    TSR    /
                                                            \         /
                                                             \       /
                                                              Model_1
}</pre>
*/
public class ShowNestedModels_36
{
   public static void main(String[] args)
   {
      // Build the Scene graph from the bottom up.

      // Create a single instance of Model_1.
      final Model m1 = new Model_1();
      ModelShading.setColor(m1, Color.red);

      // Create matricies for the nested Positions.
      final Matrix mat1 = Matrix.translate(-2, -2, 0);
      final Matrix mat2 = Matrix.translate(2, 2, 0)
                  .times( Matrix.rotateZ(180) );
      final Matrix mat5 = Matrix.translate(1, -2-Math.sqrt(2), 0)
                  .times( Matrix.scale(0.5, 0.5, 1) )
                  .times( Matrix.rotateZ(-45) );

      // Create two Positions to each refer to the model.
      final Position p4 = new Position(m1, "p4");
      final Position p5 = new Position(m1, "p5", mat5);

      // Create a Position to hold the previous two Positions
      final Position p3 = new Position(new Model("empty"), "p3");
      p3.addNestedPosition(p4, p5);

      // Create two Positions that each hold a reference to the previous Position.
      final Position p1 = new Position(new Model("empty"), "p1", mat1);
      final Position p2 = new Position(new Model("empty"), "p2", mat2);
      p1.addNestedPosition(p3);
      p2.addNestedPosition(p3);

      // Create a Position that holds the complex hierarchical
      // structure that draws four copies of Model_1. Add to it
      // the previous two positions.
      final Position complex_p = new Position(new Model("empty"),
                                              "complex_p");
      complex_p.addNestedPosition(p1, p2);

      // Create the top level Position.
      final Position p0 = new Position(new Model("empty"), "p0");

      // Add eight nested Position objects to the top level Position,
      // each containing a reference to a complex position structure
      // that is rotated around the perimeter of the image.
      for (int i = 0; i < 8; ++i)
      {
         final Matrix mat = Matrix.rotateZ(i*45)
                     .times(Matrix.translate(0, -11, 0))
                     .times(Matrix.scale(0.5));
         final Position p = new Position(new Model("empty"),
                                         "n"+i,
                                         mat);
         p.addNestedPosition( complex_p );
         p0.addNestedPosition( p );
      }

      // Add a ninth nested Position object to the top level Position.
      // This one contains a reference to a complex position structure
      // that is centered in the image.
      final Position p = new Position( new Model("empty"), "n9" );
      p.addNestedPosition( complex_p );
      p0.addNestedPosition( p );

      // Create the Scene object.
      final Scene scene = new Scene("ShowNestedModels_36",
                                    Camera.projOrtho(-15, +15, -15, +15));

      // Add the top level Position to the Scene.
      scene.addPosition(p0);


      DrawSceneGraph.draw(scene, "SG_ShowNestedModels_36");
      DrawSceneGraphTopologicalSort.draw(scene, "SG_ShowNestedModels_36_TS");


      // Create a FrameBuffer to render our Scene into.
      final int width  = 1024;
      final int height = 1024;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.white);

      //PipelineLogger.debug = true;

      for (int i = 0; i <= 36; ++i)
      {
         scene.setPosition(0,
                  scene.getPosition(0).transform( Matrix.rotateZ(10*i) ));

         // What does this do?
       //complex_p.matrix2Identity();
       //complex_p.matrix.mult(rotateZ(10*i));

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_ShowNestedModels_36_Frame%02d.ppm", i));
      }
   }
}
