/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.primitives.*;
import renderer.scene.util.DrawSceneGraph;
import renderer.scene.util.DrawSceneGraphTopologicalSort;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   This is a simple hierarchical scene made up of a
   triangle with another triangle attached to each vertex.
<p>
   Here is a sketch of the scene graph for this example.
<pre>{@code
           Scene
          /     \
         /       \
    Camera     List<Position>
                   |
                   |
               Position
              /    |    \
             /     }     \
            /      |      \
      Matrix     Model     List<Position>
       RT     (triangle1) /     |        \
                         /      |         \
                        /       |          \
                       /        |           \
                      /         |            \
              Position      Position       Position
             /     \        /    \           /   \
            /       \      /      \         /     \
      Matrix    List<   Matrix  List<     Matrix   List<
        T    Position>    TR    Position>   TR     Position>
                      \            |             /
                        \          |           /
                          \        |         /
                            \      |       /
                              \    |     /
                                Position
                              /   |    \
                             /    |     \
                        Matrix   Model   List<Position>
                          R   (triangle2)    (empty)
}</pre>
*/
public class NestedModels_v2
{
   public static void main(String[] args)
   {
      // Create a Model for the top level position.
      final Model triangle1 = new Model("triangle_1");
      final double sin2PIover3 = Math.sin(2*Math.PI/3);
      final Vertex v0 = new Vertex( 1,        0,       0);
      final Vertex v1 = new Vertex(-0.5,  sin2PIover3, 0);
      final Vertex v2 = new Vertex(-0.5, -sin2PIover3, 0);
      triangle1.addVertex(v0, v1, v2);
      triangle1.addPrimitive(new LineSegment(0, 1),
                             new LineSegment(1, 2),
                             new LineSegment(2, 0));
      ModelShading.setColor(triangle1, Color.black);

      // Create another triangle Model for the three
      // corners of the the top level triangle Model.
      final Model triangle2 = new Model("triangle_2");
      triangle2.addVertex(v0, v1, v2);
      triangle2.addPrimitive(new LineSegment(0, 1),
                             new LineSegment(1, 2),
                             new LineSegment(2, 0));
      ModelShading.setColor(triangle2, Color.red);

      final Matrix mat1 = Matrix.translate( 1.5,   0,               0),
                   mat2 = Matrix.translate(-0.75,  1.5*sin2PIover3, 0).times(Matrix.rotateZ(120)),
                   mat3 = Matrix.translate(-0.75, -1.5*sin2PIover3, 0).times(Matrix.rotateZ(240));

      // Create a floor Model.
      final Model floor = new PanelXZ(-4, 4, -4, 4);
      ModelShading.setColor(floor, Color.black);

      // Create a framebuffer to render our scene into.
      final int width  = 1024;
      final int height = 1024;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.lightGray);

      //PipelineLogger.debug = true;

      for (int i = 0; i <= 72; ++i)
      {
         // Build the Scene from the bottom up.

         // Create a Position for triangle2.
         final Position p4 = new Position(triangle2, "p4",
                                          Matrix.rotateX(5*i));
         // Create three nested positions placed at the three
         // corners of the top level position's triangle.
         final Position p1 = new Position(new Model("empty"), "p1", mat1),
                        p2 = new Position(new Model("empty"), "p2", mat2),
                        p3 = new Position(new Model("empty"), "p3", mat3);
         // Add triangle2 at each of these three positions.
         p1.addNestedPosition(p4);
         p2.addNestedPosition(p4);
         p3.addNestedPosition(p4);
         // Create the top level Position.
         final Position p0 = new Position(triangle1,
                                          "p0",
                                          Matrix.rotateZ(5*i)
                                   .times(Matrix.translate(2, 0, 0)));
         // Put these three nested Positions into the top level Position.
         p0.addNestedPosition(p1, p2, p3);

         // Create the Scene object.
         final Scene scene = new Scene("NestedModels_v2",
                                       Camera.projPerspective()
                                             .translate(0, 0, 5));
         // Add the top level Position to the Scene.
         scene.addPosition(p0);
         // Add the floor model to the scene.
         scene.addPosition(new Position(floor,
                                        Matrix.translate(0, -4, 0)));
         // Draw one Scene graph image.
         if (0 == i)
         {
            DrawSceneGraph.draw(scene, "SG_NestedModels_v2");
            DrawSceneGraphTopologicalSort.draw(scene, "SG_NestedModels_v2_TS");
         }

         // Render
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_NestedModels_v2_Frame%02d.ppm", i));
      }
   }
}
