/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

import renderer.scene.*;
import renderer.scene.util.DrawSceneGraph;
import renderer.scene.util.DrawSceneGraphTopologicalSort;
import renderer.scene.util.ModelShading;
import renderer.models_L.*;
import renderer.pipeline.*;
import renderer.framebuffer.*;

import java.awt.Color;

/**
   Here is a sketch of this program's scene graph.
   Only the TriangularPrism model holds any geometry.
   All of the other nodes hold only a matrix.
<pre>{@code
           Scene
          /     \
         /       \
    Camera     List<Position>
                   |
                   |
               Position
              /    |    \
             /     |     \
       Matrix    Model    List<Position>
         R      (empty)      / |  | \
                            /  |  |  \
                          p1  p2  p3  p4
                            \  |  |  /
                             \ |  | /
                          TriangularPrism
}</pre>
*/
public class NestedPrisms_v1
{
   public static void main(String[] args)
   {
      final double sqrt3 = Math.sqrt(3.0);

      final Scene scene = new Scene("Prisms_v1",
                                    Camera.projPerspective()
                                          .translate(0, 0, 8));

      // Create the top level Position.
      final Position p0 = new Position(new Model("empty"), "p0");

      // Add the top level Position to the Scene.
      scene.addPosition(p0);

      // Create four nested Positions each holding
      // a reference to a shared prism Model.
      final Model prism = new TriangularPrism(1.0/sqrt3, 2.0, Math.PI/4.0, 25);
      ModelShading.setColor(prism, Color.red);
      final Position p1 = new Position(prism, "p1", Matrix.translate(2+0.5/sqrt3, 0, 0));
      final Position p2 = new Position(prism, "p2", Matrix.translate(-2-0.5/sqrt3, 0, 0)
                                             .times(Matrix.rotateZ(180)));
      final Position p3 = new Position(prism, "p3", Matrix.rotateZ(90)
                                             .times(Matrix.translate(2+0.5/sqrt3, 0, 0)));
      final Position p4 = new Position(prism, "p4", Matrix.rotateZ(-90)
                                             .times(Matrix.translate(2+0.5/sqrt3, 0, 0)));

      // Put these four nested Positions into the top level Position.
      p0.addNestedPosition(p1, p2, p3, p4);

      // Create a floor Model.
      final Model floor = new PanelXZ(-4, 4, -4, 4);
      ModelShading.setColor(floor, Color.black);
      final Position floor_p = new Position(floor,
                                            "floor",
                                            Matrix.translate(0, -4, 3));
      scene.addPosition(floor_p);

      DrawSceneGraph.draw(scene, "SG_NestedPrisms_v1");
      DrawSceneGraphTopologicalSort.draw(scene, "SG_NestedPrisms_v1_TS");


      // Create a framebuffer to render our scene into.
      final int width  = 1024;
      final int height = 1024;
      final FrameBuffer fb = new FrameBuffer(width, height, Color.white);

      //PipelineLogger.debug = true;

      // Spin the model 360 degrees around three axes.
      for (int i = 0; i <= 180; ++i)
      {
         scene.setPosition(0,
                  scene.getPosition(0).transform(Matrix.rotateX(2*i)
                                          .times(Matrix.rotateY(2*i))
                                          .times(Matrix.rotateZ(2*i))));

         // Render again.
         fb.clearFB();
         Pipeline.render(scene, fb);
         fb.dumpFB2File(String.format("PPM_NestedPrism_v1_Frame_%03d.ppm", i));
      }
   }
}
