/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.util.CheckModel;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our eighth rendering pipeline. This renderer
   implements "hierarchical scenes" by recursively traversing the
   DAG of nested {@link Position}s below each of a {@link Scene}'s
   {@link Position} objects. As the renderer traverses deeper into
   the DAG of nested {@link Position}s, it accumulates a "current
   transformation matrix" that transforms each {@link Vertex} from a
   {@link Model}'s local coordinate system to the (shared) world
   coordinate system. The recursive traversal of the scene's DAG is
   not a new pipeline stage, so there are still seven pipeline stages.
*/
public final class Pipeline
{
   // Mostly for compatibility with renderers 1 through 3.
   public static Color DEFAULT_COLOR = Color.white;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      PipelineLogger.debugScene = scene.debug;

      logMessage("\n== Begin Rendering of Scene (Pipeline 1): " + scene.name + " ==");

      logMessage("-- Current Camera:\n" + scene.camera);

      // For every Position in the Scene, render the Position's Model.
      for (final Position position : scene.positionList)
      {
         PipelineLogger.debugPosition = position.debug;

         if ( position.visible )
         {
            // Begin a pre-order, depth-first-traversal from this Position.
            render_position(scene, position, Matrix.identity(), vp);
         }
         else
         {
            logMessage("==== Hidden position: " + position.name);
         }
      }
      logMessage("== End Rendering of Scene (Pipeline 1) ==");
   }


   /**
      Recursively renderer a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then renders the {@link Model} in {@code position}
      using the updated {@code ctm} in the {@link Model2View} stage.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively render
      @param ctm       current model-to-view transformation {@link Matrix}
      @param vp       {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   private static void render_position(final Scene scene,
                                       final Position position,
                                       final Matrix ctm,
                                       final FrameBuffer.Viewport vp)
   {
      // Opening message.
      logMessage("==== Render position: " + position.name + " ====");

      logMessage("---- Transformation matrix:\n" + position.matrix);

      // Update the current model-to-view transformation matrix.
      final Matrix ctm2 = ctm.times( position.matrix );

      // Render this Position's Model, if the model is visible.
      if ( position.model.visible )
      {
         logMessage("====== Render model: " + position.model.name + " ======");

         CheckModel.check(position.model);

         // Mostly for compatibility with renderers 1 through 3.
         if (  position.model.colorList.isEmpty()
           && !position.model.vertexList.isEmpty())
         {
            for (int i = 0; i < position.model.vertexList.size(); ++i)
            {
               position.model.addColor( DEFAULT_COLOR );
            }
            System.err.println("***WARNING: Added default color to model: "
                              + position.model.name + ".");
         }

         logVertexList("0. Model        ", position.model);

         // 1. Apply the current model-to-world coordinate transformation.
         final Model model1 = Model2World.model2world(position, ctm2);

         logVertexList("1. World        ", model1);

         // 2. Apply the Camera's world-to-view coordinate transformation.
         final Model model2 = World2View.world2view(model1,
                                                    scene.camera);

         logVertexList("2. View         ", model2);

         // 3. Apply the Camera's normalizing view-to-camera coordinate transformation.
         final Model model3 = View2Camera.view2camera(model2, scene.camera);

         logVertexList("3. Camera       ", model3);
         logColorList("3. Camera       ", model3);
         logPrimitiveList("3. Camera       ", model3);

         // 4. Clip primitives to the camera's near plane.
         final Model model4 = NearClip.clip(model3, scene.camera);

         logVertexList("4. Near_Clipped", model4);
         logColorList("4. Near_Clipped", model4);
         logPrimitiveList("4. Near_Clipped", model4);

         // 5. Apply the Camera's projection transformation.
         final Model model5 = Projection.project(model4, scene.camera);

         logVertexList("5. Projected   ", model5);

         // 6. Clip primitives to the camera's view rectangle.
         final Model model6 = Clip.clip(model5);

         logVertexList("6. Clipped     ", model6);
         logColorList("6. Clipped     ", model6);
         logPrimitiveList("6. Clipped     ", model6);

         // 7. Rasterize every visible primitive into pixels.
         Rasterize.rasterize(model6, vp);

         logMessage("====== End model: "
                            + position.model.name + " ======");
      }
      else
      {
         logMessage("====== Hidden model: " + position.model.name + " ======");
      }

      // Recursively render every nested Position of this Position.
      for (final Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            // Do a pre-order, depth-first-traversal from this nested Position.
            render_position(scene, p, ctm2, vp); // recursion
         }
         else
         {
            logMessage("====== Hidden position " + position.name + " ====");
         }
      }

      // Closing message.
      logMessage("==== End position: " + position.name + " ====");
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Pipeline() {
      throw new AssertionError();
   }
}
