/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.pipeline;

import renderer.scene.*;
import renderer.scene.util.CheckModel;
import renderer.framebuffer.*;
import static renderer.pipeline.PipelineLogger.*;

import java.awt.Color;

/**
   This renderer takes as its input a {@link Scene} data structure
   and a {@link FrameBuffer.Viewport} within a {@link FrameBuffer}
   data structure. This renderer mutates the {@link FrameBuffer.Viewport}
   so that it is filled in with the rendered image of the geometric
   scene represented by the {@link Scene} object.
<p>
   This implements our eighth rendering pipeline. This renderer
   implements "hierarchical scenes" by recursively traversing the
   DAG of nested {@link Position}s below each of a {@link Scene}'s
   {@link Position} objects. As the renderer traverses deeper into
   the DAG of nested {@link Position}s, it accumulates a "current
   transformation matrix" that transforms each {@link Vertex} from a
   {@link Model}'s local coordinate system to the (shared) world
   coordinate system. The recursive traversal of the scene's DAG is
   not a new pipeline stage, so there are still seven pipeline stages.
*/
public final class Pipeline2
{
   // Mostly for compatibility with renderers 1 through 3.
   public static Color DEFAULT_COLOR = Color.white;

   // Make all the intermediate Scene objects
   // available for special effects processing.
   public static Scene scene1 = null;
   public static Scene scene2 = null;
   public static Scene scene3 = null;
   public static Scene scene4 = null;
   public static Scene scene5 = null;
   public static Scene scene6 = null;

   /**
      Mutate the {@link FrameBuffer}'s default {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param fb     {@link FrameBuffer} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer fb)
   {
      render(scene, fb.vp); // render into the default viewport
   }


   /**
      Mutate the {@link FrameBuffer}'s given {@link FrameBuffer.Viewport}
      so that it holds the rendered image of the {@link Scene} object.

      @param scene  {@link Scene} object to render
      @param vp     {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   public static void render(final Scene scene, final FrameBuffer.Viewport vp)
   {
      PipelineLogger.debugScene = scene.debug;

      logMessage("\n= Begin Rendering of Scene (Pipeline 2): " + scene.name + " =");

      logMessage("- Current Camera:\n" + scene.camera);

      scene1 = new Scene(scene.name, scene.camera);

      logMessage("== 1. Begin model-to-world transformation of Scene ====");
      for (final Position position : scene.positionList)
      {
         PipelineLogger.debugPosition = position.debug;

         if ( position.visible )
         {
            final Position renderedPosition = model2world(position, Matrix.identity());
            scene1.addPosition( renderedPosition );
         }
         else
         {
            logMessage("==== 1. Hidden position: "
                             + position.name + " ====");
         }
      }
      logMessage("== 1. End model-to-world transformation of Scene ====");

      scene2 = new Scene(scene.name, scene.camera);

      logMessage("== 2. Begin world-to-view transformation of Scene ====");
      for (final Position position : scene1.positionList)
      {
         scene2.addPosition( world2view(position, scene.camera) );
      }
      logMessage("== 2. End world-to-view transformation of Scene ====");

      scene3 = new Scene(scene.name, scene.camera);

      logMessage("== 3. Begin view-to-camera transformation of Scene ====");
      for (final Position position : scene2.positionList)
      {
         scene3.addPosition( view2camera(position, scene.camera) );
      }
      logMessage("== 3. End view-to-camera transformation of Scene ====");

      scene4 = new Scene(scene.name, scene.camera);

      logMessage("== 4. Begin near-plane clipping of Scene ====");
      for (final Position position : scene3.positionList)
      {
         scene4.addPosition( nearClip(position, scene.camera) );
      }
      logMessage("== 4. End near-plane clipping of Scene ====");

      scene5 = new Scene(scene.name, scene.camera);

      logMessage("== 5. Begin projection transformation of Scene ====");
      for (final Position position : scene4.positionList)
      {
         scene5.addPosition( project(position, scene.camera) );
      }
      logMessage("== 5. End projection transformation of Scene ====");

      scene6 = new Scene(scene.name, scene.camera);

      logMessage("== 6. Begin primitive clipping of Scene ====");
      for (final Position position : scene5.positionList)
      {
         scene6.addPosition( clip(position) );
      }
      logMessage("== 6. End primitive clipping of Scene ====");

      logMessage("== 7. Begin primitive rasterization of Scene ====");
      for (final Position position : scene6.positionList)
      {
         rasterize(position, vp);
      }
      logMessage("== 7. End primitive rasterization of Scene ====");

      logMessage("= End Rendering of Scene (Pipeline 2) =");
   }


   /**
      Pipeline stage 1, model-to-world transformation.
      <p>
      Recursively transform a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.
      <p>
      The pre-order "visit node" operation in this traversal first updates the
      "current transformation matrix", ({@code ctm}), using the {@link Matrix}
      in {@code position} and then transforms the {@link Model} in {@code position}
      using the updated {@code ctm}.

      @param scene     the {@link Scene} that we are rendering
      @param position  the current {@link Position} object to recursively transform
      @param ctm       current model-to-world transformation {@link Matrix}
      @return a tree of transformed {@link Position} objects
   */
   private static Position model2world(final Position position,
                                       final Matrix ctm)
   {
      // Opening message.
      logMessage("==== 1. Render position: " + position.name + " ====");

      logMessage("---- Transformation matrix:\n" + position.matrix);

      // Update the current model-to-world transformation matrix.
      final Matrix ctm2 = ctm.times( position.matrix );

      // Create a new Position to hold the newly rendered
      // Model and the newly rendered sub-Positions.
      final Position position2;

      // Render this Position's model, if the model is visible and not empty.
      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 1. Model-to-world transformation of: "
                            + position.model.name + " ======");

         CheckModel.check(position.model);

         // Mostly for compatibility with renderers 1 through 3.
         if (  position.model.colorList.isEmpty()
           && !position.model.vertexList.isEmpty())
         {
            for (int i = 0; i < position.model.vertexList.size(); ++i)
            {
               position.model.addColor( DEFAULT_COLOR );
            }
            System.err.println("***WARNING: Added default color to model: "
                              + position.model.name + ".");
         }

         logVertexList("0. Model    ", position.model);

         position2 = new Position(
                            Model2World.model2world(position, ctm2),
                            position.name);

         logVertexList("1. World    ", position2.model);

         logMessage("====== 1. End Model: "
                            + position2.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 1. Hidden model: "
                               + position.model.name + " ======");
         }
         else
         {
            logMessage("====== 1. Empty model ======");
         }
         position2 = new Position(position.model, position.name);
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         if ( p.visible )
         {
            final Position renderedPosition = model2world(p, ctm2); // recursion
            position2.addNestedPosition( renderedPosition );
         }
         else
         {
            logMessage("==== 1. Hidden position: " + p.name + " ====");
         }
      }

      // Closing message.
      logMessage("==== 1. End position: " + position.name + " ====");

      return position2;
   }


   /**
      Pipeline stage 2, world-to-view transformation.
      <p>
      Recursively transform a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.

      @param position  the current {@link Position} object to recursively transform
      @param camera    the {@link Scene}'s {@link Camera} with its location data
      @return a tree of transformed {@link Position} objects
   */
   private static Position world2view(final Position position,
                                      final Camera camera)
   {
      logMessage("==== 2. Render position: " + position.name + " ====");

      // Create a new Position to hold the newly rendered
      // Model and the newly rendered sub-Positions.
      final Position position2;

      // Render this Position's model, if the model is visible and not empty.
      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 2. Transform model: "
                            + position.model.name + " ======");

         position2 = new Position(
                            World2View.world2view(position.model, camera),
                            position.name);

         logVertexList("2. View     ", position2.model);

         logMessage("====== 2. End Model: "
                            + position2.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 2. Hidden model: "
                               + position.model.name + " ======");
         }
         else
         {
            logMessage("====== 2. Empty model ======");
         }
         position2 = new Position(position.model, position.name);
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         position2.addNestedPosition( world2view(p, camera) ); // recursion
      }

      logMessage("==== 2. End position: " + position.name + " ====");

      return position2;
   }


   /**
      Pipeline stage 3, view-to-camera transformation.
      <p>
      Recursively transform a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.

      @param position  the current {@link Position} object to recursively transform
      @param camera    the {@link Scene}'s {@link Camera} with the view volume data
      @return a tree of transformed {@link Position} objects
   */
   private static Position view2camera(final Position position,
                                       final Camera camera)
   {
      logMessage("==== 3. Render position: " + position.name + " ====");

      // Create a new Position to hold the newly rendered
      // Model and the newly rendered sub-Positions.
      final Position position2;

      // Render this Position's model, if the model is visible and not empty.
      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 3. Transform model: "
                            + position.model.name + " ======");

         position2 = new Position(
                            View2Camera.view2camera(position.model, camera),
                            position.name);

         logVertexList("3. Camera   ", position2.model);
         logColorList("3. Camera   ", position2.model);
         logPrimitiveList("3. Camera   ", position2.model);

         logMessage("====== 3. End Model: "
                            + position2.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 3. Hidden model: "
                               + position.model.name + " ======");
         }
         else
         {
            logMessage("====== 3. Empty model ======");
         }
         position2 = new Position(position.model, position.name);
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         position2.addNestedPosition( view2camera(p, camera) ); // recursion
      }

      logMessage("==== 3. End position: " + position.name + " ====");

      return position2;
   }


   /**
      Pipeline stage 4, near-plane clipping.
      <p>
      Recursively clip a {@link Position} at the {@link Camera}'s near plane.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.

      @param position  the current {@link Position} object to recursively clip
      @param camera  {@link Camera} that determines the near clipping plane
      @return a tree of clipped {@link Position} objects
   */
   private static Position nearClip(final Position position,
                                    final Camera camera)
   {
      logMessage("==== 4. Render position: " + position.name + " ====");

      // Create a new Position to hold the newly rendered
      // Model and the newly rendered sub-Positions.
      final Position position2;

      // Render this Position's model, if the model is visible and not empty.
      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 4. Near_Clip model: "
                            + position.model.name + " ======");

         position2 = new Position(
                            NearClip.clip(position.model, camera),
                            position.name);

         logVertexList("4. Near_Clipped  ", position2.model);
         logColorList("4. Near_Clipped  ", position2.model);
         logPrimitiveList("4. Near_Clipped  ", position2.model);

         logMessage("====== 4. End Model: "
                            + position2.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 4. Hidden model: "
                               + position.model.name + " ======");
         }
         else
         {
            logMessage("====== 4. Empty model ======");
         }
         position2 = new Position(position.model, position.name);
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         position2.addNestedPosition( nearClip(p, camera) ); // recursion
      }

      logMessage("==== 4. End position: " + position.name + " ====");

      return position2;
   }


   /**
      Pipeline stage 5, projection transformation.
      <p>
      Recursively project a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.

      @param position  the current {@link Position} object to recursively project
      @param camera    the {@link Scene}'s {@link Camera} with the view volume data
      @return a tree of projected {@link Position} objects
   */
   private static Position project(final Position position,
                                   final Camera camera)
   {
      logMessage("==== 5. Render position: " + position.name + " ====");

      // Create a new Position to hold the newly rendered
      // Model and the newly rendered sub-Positions.
      final Position position2;

      // Render this Position's model, if the model is visible and not empty.
      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 5. Project model: "
                            + position.model.name + " ======");

         position2 = new Position(
                            Projection.project(position.model, camera),
                            position.name);

         logVertexList("5. Projected", position2.model);

         logMessage("====== 5. End Model: "
                            + position2.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 5. Hidden model: "
                               + position.model.name + " ======");
         }
         else
         {
            logMessage("====== 5. Empty model ======");
         }
         position2 = new Position(position.model, position.name);
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         position2.addNestedPosition( project(p, camera) ); // recursion
      }

      logMessage("==== 5. End position: " + position.name + " ====");

      return position2;
   }


   /**
      Pipeline stage 6, clipping.
      <p>
      Recursively clip a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.

      @param position  the current {@link Position} object to recursively clip
      @return a tree of clipped {@link Position} objects
   */
   private static Position clip(final Position position)
   {
      logMessage("==== 6. Render position: " + position.name + " ====");

      // Create a new Position to hold the newly rendered
      // Model and the newly rendered sub-Positions.
      final Position position2;

      // Render this Position's model, if the model is visible and not empty.
      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 6. Clip model: "
                            + position.model.name + " ======");

         position2 = new Position(
                            Clip.clip(position.model),
                            position.name);

         logVertexList("6. Clipped  ", position2.model);
         logColorList("6. Clipped  ", position2.model);
         logPrimitiveList("6. Clipped  ", position2.model);

         logMessage("====== 6. End Model: "
                            + position2.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 6. Hidden model: "
                               + position.model.name + " ======");
         }
         else
         {
            logMessage("====== 6. Empty model ======");
         }
         position2 = new Position(position.model, position.name);
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         position2.addNestedPosition( clip(p) ); // recursion
      }

      logMessage("==== 6. End position: " + position.name + " ====");

      return position2;
   }


   /**
      Pipeline stage 7, rasterization.
      <p>
      Recursively rasterize a {@link Position}.
      <p>
      This method does a pre-order, depth-first-traversal of the tree of
      {@link Position}'s rooted at the parameter {@code position}.

      @param position  the current {@link Position} object to recursively rasterize
      @param vp       {@link FrameBuffer.Viewport} to hold rendered image of the {@link Scene}
   */
   private static void rasterize(final Position position,
                                 final FrameBuffer.Viewport vp)
   {
      logMessage("==== 7. Render position: " + position.name + " ====");

      if ( position.model.visible && !position.model.vertexList.isEmpty() )
      {
         logMessage("====== 7. Rasterize model: "
                            + position.model.name + " ======");

         Rasterize.rasterize(position.model, vp);

         logMessage("====== 7. End Model: "
                            + position.model.name + " ======");
      }
      else
      {
         if ( ! position.model.visible )
         {
            logMessage("====== 7. Hidden model: "
                               + position.name + " ======");
         }
         else
         {
            logMessage("====== 7. Empty model ======");
         }
      }

      // Do a pre-order, depth-first-traversal from this nested Position.
      for (final Position p : position.nestedPositions)
      {
         rasterize(p, vp); // recursion
      }

      logMessage("==== 7. End position: " + position.name + " ====");
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private Pipeline2() {
      throw new AssertionError();
   }
}
