/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene.util;

import renderer.scene.*;

import java.io.File;
import java.io.PrintWriter;
import java.lang.Runtime;

/**
   This program converts a {@link Scene} data structure into
   a DOT description of the scene. The DOT description is
   then processed by the dot.exe program to produce a
   graphical image of the scene data structure.
<p>
   Create a <code>scene.png</code> file from a <code>scene.dot</code>
   file with the following command line.
   <pre>{@code
   > dot.exe -Tpng -O scene.dot
   }</pre>
<p>
   See
<br><a href="https://www.graphviz.org/Documentation.php" target="_top">
             https://www.graphviz.org/Documentation.php</a>
*/
public class DrawSceneGraph
{
   // These two variables are the addresses of a node
   // in the topological sort of the scene graph DAG.
   // They are used to detect those nodes that have
   // multiple parents.
   private static int positionNumber;
   private static int modelNumber;

   /**
      This method converts the {@link Scene} data structure into a dot
      language description, then writes the dot code into a dot file,
      then runs the dot.exe program on the dot file to create a png
      image of the scene data structure.

      @param scene     {@link Scene} that needs to be converted to a dot description
      @param fileName  base name for the dot and png files
   */
   public static void draw(final Scene scene, final String fileName)
   {
      // Convert the scene data structure into a dot language description.
      final String dotDescription = scene2dot(scene);

      // Write the dot language description stored in dotDescription
      // into a dot file. Then use the dot.exe program to convert the
      // dot file into a png file.
      try
      {
         // Create the (empty) dot file.
         String baseName = fileName;
         java.io.PrintWriter out = new PrintWriter(
                                      new File(baseName + ".dot"));

         // Write the dot commands into the dot file.
         out.print( dotDescription );
         out.close();

         // Create a command-line for running the dot.exe program.
         final String dotExecutable = "C:\\Graphviz\\bin\\dot.exe";
         String[] cmd = {dotExecutable,
                         "-Tpng",
                         baseName + ".dot",
                         "-o",
                         baseName + ".png"};

         File dot = new File(dotExecutable);
         if(dot.exists() && !dot.isDirectory())
         {
            // Execute the command-line to create the png file.
            Runtime.getRuntime().exec(cmd);
         }
         else
         {
            System.out.println("\nPlease consider installing GraphViz:");
            System.out.println("  https://graphviz.org/download/");
            System.out.println("or upload the contents of " + baseName + ".dot to Graphviz Visual Editor:");
            System.out.println("  http://magjac.com/graphviz-visual-editor/");
         }
      }
      catch (Exception e)
      {
         System.out.println( e );
      }
   }


   /**
      This method generates a dot language description of the
      DAG rooted at {@code scene}.
      <p>
      This method generates the dot code for the forest of top-level
      positions just below the scene node. Each position node just
      below the scene node is the root of a DAG. This method calls
      the <code>position2dot()</code> method to recursively traverse
      the DAG of each top-level position. Every position has attached
      to it a {@link Model}, so every position node has attached to it
      a model node.

      @param  scene {@link Scene} that needs to be converted to a dot description
      @return a {@link String} containing the dot language description of the scene
   */
   public static String scene2dot(final Scene scene)
   {
      // https://stackoverflow.com/questions/48266439/getting-graphviz-to-eliminate-identical-duplicate-edges
      String result = "strict digraph {\n";

      // https://graphviz.org/docs/attrs/ordering/
      result += "graph [ordering=\"out\"];\n";

      // https://stackoverflow.com/questions/10879115/graphviz-change-font-for-the-whole-graph
      // https://graphviz.org/docs/attrs/fontname/
      result += "graph [fontname=\"helvetica\"];\n";
      result += "node  [fontname=\"helvetica\"];\n";
      result += "edge  [fontname=\"helvetica\"];\n";

      // Scene node.
      result += "scene [label=\"Scene: " + scene.name + "\"];\n";

      // Camera and List<Position> nodes under the Scene node.

      // Camera node and label.
      final String cameraNodeName = "Camera";
      result += cameraNodeName + " ";
      result += "[label=\"" + scene.camera + "\"];\n";
      // Camera edge.
      result += "scene -> " + cameraNodeName + ";\n";

      // List<Position> node and label.
      final String pListNodeName = "positionList";
      result += pListNodeName + " ";
      result += "[label=\"List<Position>\"];\n";
      // List<Position> edge.
      result += "scene -> " + pListNodeName + ";\n";

      positionNumber = -1;
      modelNumber = -1;

      // Create a node and an edge for each top-level
      // Position and its Model.
      for (int i = 0; i < scene.positionList.size(); ++i)
      {
         ++positionNumber;

         // Position node name.
         final Position positionReference = scene.getPosition(i);
         final int positionAddress = computePositionAddress(scene, positionReference);
         final String pNodeName = "_p" + positionAddress;

         // Position node and label.
         result += pNodeName + " ";
         result += "[label=\"Position: " + positionReference.name + "\"];\n";

         // Position edge.
         result += pListNodeName + " -> " + pNodeName + ";\n";

         // Model node name and label.
         ++modelNumber;
         final Model modelReference = positionReference.model;
         final String mNodeName;

         // Check if the Model is being reused.
         final int modelAddress = computeModelAddress(scene, modelReference);
         // Model node name.
         mNodeName = "_m" + modelAddress;
         // Model node and label (if the Model hasn't already been used).
         if (modelAddress == modelNumber)
         {
            result += mNodeName + " ";
            result += "[label=\"Model: " + modelReference.name + "\"];\n";
         }

         // Model edge.
         result += pNodeName + " -> " + mNodeName + ";\n";

         // Convert this position's nested positions into a dot description.
         if ( ! positionReference.nestedPositions.isEmpty() )
         {
            // positionList node and label.
            result += pNodeName + "_List ";
            result += "[label=\"List<Position>\"];\n";
            // positionList edge.
            result += pNodeName + " -> " + pNodeName + "_List;\n";
            // Nested positions.
            result += position2dot(scene, positionReference, pNodeName + "_List");
         }
      }

      result += "}\n";

      return result;
   }


   /**
      This <code>position2dot()</code> method is essentially a
      preorder traversal of the DAG of {@link Position} nodes.
      <p>
      This method generates a dot language description of the
      DAG rooted at the give {@code position}.
      <p>
      {@code nodeLabel} is the id that has been assigned to the
      node for the given {@code position}. Every position has
      attached to it a {@link Model}, so every position node
      has attached to it a model node.

      @param position  {@link Position} that needs to be converted to a dot description
      @param nodeName  the {@link String} name that has been assigned to {@code position}
      @return a {@link String} containing the dot language description of the position
   */
   public static String position2dot(final Scene scene,
                                     final Position position,
                                     final String nodeName)
   {
      String result = "";

      // Create nodes and edges for each nested position.
      for (int i = 0; i < position.nestedPositions.size(); ++i)
      {
         ++positionNumber;

         // Position node name.
         final Position positionReference = position.getNestedPosition(i);
         // Check if the Position is being reused.
         final int positionAddress = computePositionAddress(scene, positionReference);
         final String newNodeName = "_p" + positionAddress;

         // Position node and label (if this Position hasn't already been used).
         if (positionAddress == positionNumber)
         {
            result += newNodeName + " ";
            result += "[label=\"Position: " + positionReference.name + "\"];\n";
         }

         // Position edge.
         result += nodeName + " -> " + newNodeName + ";\n";

         // Model node name and label.
         ++modelNumber;
         final Model modelReference = positionReference.model;

         // Check if the Model is being reused.
         final int modelAddress = computeModelAddress(scene, modelReference);
         // Model node name.
         final String mNodeName = "_m" + modelAddress;
         // Model node and label (if the Model hasn't already been used).
         if (modelAddress == modelNumber)
         {
            result += mNodeName + " ";
            result += "[label=\"Model: " + modelReference.name + "\"];\n";
         }

         // Model edge.
         result += newNodeName + " -> " + mNodeName + ";\n";

         // Recursively convert this position's nested positions into a dot description.
         if ( ! positionReference.nestedPositions.isEmpty() )
         {
            // positionList node and label.
            result += newNodeName + "_List ";
            result += " [label=\"List<Position>\"];\n";
            // positionList edge.
            result += newNodeName + " -> " + newNodeName + "_List;\n";
            // Nested positions.
            result += position2dot(scene, positionReference, newNodeName + "_List");
         }
      }

      return result;
   }


   private static int computeModelAddress;
   private static int computeModelAddress(Scene scene, Model model)
   {
      computeModelAddress = -1;
      for (Position p : scene.positionList)
      {
         final int rv = computeModelAddressHelper(p, model);
         if (rv >= 0) return rv;
      }

      throw new IllegalArgumentException("model not found in the scene");
   }

   private static int computeModelAddressHelper(Position p, Model model)
   {
      ++computeModelAddress;
      if ( model == p.model )
      {
         return computeModelAddress;
      }
      for (Position p2 : p.nestedPositions)
      {
         final int rv = computeModelAddressHelper(p2, model);
         if (rv >= 0) return rv;
      }
      return -1;
   }


   private static int computePositionAddress;
   private static int computePositionAddress(Scene scene, Position target)
   {
      computePositionAddress = -1;
      for (Position p : scene.positionList)
      {
         final int rv = computePositionAddressHelper(p, target);
         if (rv >= 0) return rv;
      }

      throw new IllegalArgumentException("position not found in the scene");
   }

   private static int computePositionAddressHelper(Position position, Position target)
   {
      ++computePositionAddress;
      if ( target == position )
      {
         return computePositionAddress;
      }
      for (Position p : position.nestedPositions)
      {
         final int rv = computePositionAddressHelper(p, target);
         if (rv >= 0) return rv;
      }
      return -1;
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private DrawSceneGraph()
   {
      throw new AssertionError();
   }
}
