/*
 * Renderer 10. The MIT License.
 * Copyright (c) 2022 rlkraft@pnw.edu
 * See LICENSE for details.
*/

package renderer.scene.util;

import renderer.scene.*;

import java.io.File;
import java.io.PrintWriter;
import java.lang.Runtime;

/**
   This program converts a {@link Scene} data structure into
   a DOT description of the scene. The DOT description is
   then processed by the dot.exe program to produce a
   graphical image of the scene data structure.
<p>
   Create a <code>scene.png</code> file from a <code>scene.dot</code>
   file with the following command line.
   <pre>{@code
   > dot.exe -Tpng -O scene.dot
   }</pre>
<p>
   See
<br><a href="https://www.graphviz.org/Documentation.php" target="_top">
             https://www.graphviz.org/Documentation.php</a>
*/
public class DrawSceneGraphTopologicalSort
{
   /**
      This method converts the {@link Scene} data structure into a dot
      language description, then writes the dot code into a dot file,
      then runs the dot.exe program on the dot file to create a png
      image of the scene data structure.

      @param scene     {@link Scene} that needs to be converted to a dot description
      @param fileName  base name for the dot and png files
   */
   public static void draw(final Scene scene, final String fileName)
   {
      // Convert the scene data structure into a dot language description.
      final String dotDescription = scene2dot(scene);

      // Write the dot language description stored in dotDescription
      // into a dot file. Then use the dot.exe program to convert the
      // dot file into a png file.
      try
      {
         // Create the (empty) dot file.
         String baseName = fileName;
         java.io.PrintWriter out = new PrintWriter(
                                      new File(baseName + ".dot"));

         // Write the dot commands into the dot file.
         out.print( dotDescription );
         out.close();

         // Create a command-line for running the dot.exe program.
         String[] cmd = {"C:\\graphviz-2.38\\release\\bin\\dot.exe",
                         "-Tpng",
                         baseName + ".dot",
                         "-o",
                         baseName + ".png"};

         File dot = new File("C:\\graphviz-2.38\\release\\bin\\dot.exe");
         if(dot.exists() && !dot.isDirectory())
         {
            // Execute the command-line to create the png file.
            Runtime.getRuntime().exec(cmd);
         }
         else
         {
            System.out.println("\nPlease consider installing GraphViz:");
            System.out.println("  https://graphviz.org/download/");
            System.out.println("or upload the contents of " + baseName + ".dot to Graphviz Visual Editor:");
            System.out.println("  http://magjac.com/graphviz-visual-editor/");
         }
      }
      catch (Exception e)
      {
         System.out.println( e );
      }
   }


   /**
      This method generates a dot language description of the
      DAG rooted at {@code scene}.
      <p>
      This method generates the dot code for the forest of top-level
      positions just below the scene node. Each position node just
      below the scene node is the root of a DAG. This method calls
      the <code>position2dot()</code> method to recursively traverse
      the DAG of each top-level position. Every position has attached
      to it a {@link Model}, so every position node has attached to it
      a model node.

      @param  scene {@link Scene} that needs to be converted to a dot description
      @return a {@link String} containing the dot language description of the scene
   */
   public static String scene2dot(final Scene scene)
   {
      String result = "digraph {\n";

      // https://stackoverflow.com/questions/10879115/graphviz-change-font-for-the-whole-graph
      // https://graphviz.org/docs/attrs/fontname/
      result += "graph [fontname=\"helvetica\"];\n";
      result += "node  [fontname=\"helvetica\"];\n";
      result += "edge  [fontname=\"helvetica\"];\n";

      // Scene node.
      result += "scene [label=\"Scene: " + scene.name + "\"];\n";

      // Camera and List<Position> nodes under the Scene node.

      // Camera node and label.
      final String cameraNodeName = "Camera";
      result += cameraNodeName + " ";
      result += "[label=\"Camera\"];\n";
      // Camera edge.
      result += "scene -> " + cameraNodeName + ";\n";

      // List<Position> node and label.
      final String pListNodeName = "positionList";
      result += pListNodeName + " ";
      result += "[label=\"List<Position>\"];\n";
      // List<Position> edge.
      result += "scene -> " + pListNodeName + ";\n";

      // Create a node and an edge for each top-level Position and its Model.
      for (int i = 0; i < scene.positionList.size(); ++i)
      {
         // Position node name.
         final String pNodeName = "_p" + i;

         // Position node and label.
         result += pNodeName + " ";
         result += "[label=\"Position: " + scene.getPosition(i).name + "\"];\n";

         // Position edge.
         result += pListNodeName + " -> " + pNodeName + ";\n";

         // Model node name.
         final String mNodeName = "_m" + i;

         // Model node.
         result += mNodeName + " ";

         // Model node label.
         result += "[label=\"Model: " + scene.getPosition(i).model.name + "\"];\n";

         // Model edge.
         result += pNodeName + " -> " + mNodeName + ";\n";

         // Convert this position's nested positions into a dot description.
         if ( ! scene.getPosition(i).nestedPositions.isEmpty() )
         {
            // positionList node and label.
            result += pNodeName + "_List ";
            result += "[label=\"List<Position>\"];\n";
            // positionList edge.
            result += pNodeName + " -> " + pNodeName + "_List;\n";
            // Nested positions.
            result += position2dot(scene.getPosition(i), pNodeName + "_List");
         }
      }

      result += "}\n";

      return result;
   }


   /**
      This <code>position2dot()</code> method is essentially a
      preorder traversal of the DAG of {@link Position} nodes.
      <p>
      This method generates a dot language description of the
      DAG rooted at the give {@code position}.
      <p>
      {@code nodeLabel} is the id that has been assigned to the
      node for the given {@code position}. Every position has
      attached to it a {@link Model}, so every position node
      has attached to it a model node.

      @param position   {@link Position} that needs to be converted to a dot description
      @param nodeName  the {@link String} name that has been assigned to {@code position}
      @return a {@link String} containing the dot language description of the position
   */
   public static String position2dot(final Position position, final String nodeName)
   {
      String result = "";

      // Create node and an edge for each nested Position.
      for (int i = 0; i < position.nestedPositions.size(); ++i)
      {
         // Position node name.
         final String newNodeName = nodeName + "_p" + i;

         // Position node and label.
         result += newNodeName + " ";
         result += "[label=\"Position: " + position.getNestedPosition(i).name + "\"];\n";

         // Position edge.
         result += nodeName + " -> " + newNodeName + ";\n";

         // Model node name.
         final String newModelName = newNodeName + "_m" + i;

         // Model node.
         result += newModelName + " ";

         // Model label.
         result += "[label=\"Model: " + position.getNestedPosition(i).model.name + "\"];\n";

         // Model edge.
         result += newNodeName + " -> " + newModelName + ";\n";

         // Recursively convert this position's nested positions into a dot description.
         if ( ! position.getNestedPosition(i).nestedPositions.isEmpty() )
         {
            // positionList node and label.
            result += newNodeName + "_List ";
            result += "[label=\"List<Position>\"];\n";
            // positionList edge.
            result += newNodeName + " -> " + newNodeName + "_List;\n";
            // Nested positions.
            result += position2dot(position.getNestedPosition(i), newNodeName + "_List");
         }
      }

      return result;
   }



   // Private default constructor to enforce noninstantiable class.
   // See Item 4 in "Effective Java", 3rd Ed, Joshua Bloch.
   private DrawSceneGraphTopologicalSort()
   {
      throw new AssertionError();
   }
}
